﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2010 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using alglib;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using System.Collections.Generic;
using HeuristicLab.Problems.DataAnalysis.Evaluators;
using HeuristicLab.Analysis;

namespace HeuristicLab.Problems.DataAnalysis.Operators {
  [Item("Covariant Parsimony Pressure Adder", "Covariant Parsimony Pressure Adder.")]
  [StorableClass]
  public class CovariantParsimonyPressureAdder : SingleSuccessorOperator {
    public IScopeTreeLookupParameter<SymbolicExpressionTree> ParentsParameter {
      get { return (IScopeTreeLookupParameter<SymbolicExpressionTree>)Parameters["Parents"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> ParentQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["ParentQuality"]; }
    }
    public IScopeTreeLookupParameter<DoubleValue> ParentAdjustedQualityParameter {
      get { return (IScopeTreeLookupParameter<DoubleValue>)Parameters["ParentAdjustedQuality"]; }
    }
    public ILookupParameter<SymbolicExpressionTree> OffspringParameter {
      get { return (ILookupParameter<SymbolicExpressionTree>)Parameters["Offspring"]; }
    }
    public ILookupParameter<DoubleValue> OffspringQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["OffspringQuality"]; }
    }
    public ILookupParameter<DoubleValue> OffspringAdjustedQualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["OffspringAdjustedQuality"]; }
    }
    public IValueLookupParameter<DoubleValue> KParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["K"]; }
    }
    public IValueLookupParameter<BoolValue> ApplyParsimonyPressureParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["ApplyParsimonyPressure"]; }
    }
    public ILookupParameter<DoubleValue> CParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["C"]; }
    }

    [StorableConstructor]
    public CovariantParsimonyPressureAdder(bool deserializing) : base(deserializing) { }
    public CovariantParsimonyPressureAdder()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<SymbolicExpressionTree>("Parents"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentQuality"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentAdjustedQuality"));
      Parameters.Add(new LookupParameter<SymbolicExpressionTree>("Offspring"));
      Parameters.Add(new LookupParameter<DoubleValue>("OffspringQuality"));
      Parameters.Add(new LookupParameter<DoubleValue>("OffspringAdjustedQuality"));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("K", new DoubleValue(1)));
      Parameters.Add(new LookupParameter<DoubleValue>("C"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("ApplyParsimonyPressure"));
    }

    [StorableHook(Persistence.Default.CompositeSerializers.Storable.HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }

    public override IOperation Apply() {
      if (ApplyParsimonyPressureParameter.ActualValue.Value) {
        double k = KParameter.ActualValue.Value;
        double c = CParameter.ActualValue.Value;

        // handle offspring
        var offspringLength = OffspringParameter.ActualValue.Size;
        OffspringAdjustedQualityParameter.ActualValue = new DoubleValue(OffspringQualityParameter.ActualValue.Value - c * Math.Pow(offspringLength, k));

        // handle parents
        ItemArray<SymbolicExpressionTree> parents = ParentsParameter.ActualValue;
        ItemArray<DoubleValue> parentQualities = ParentQualityParameter.ActualValue;

        ItemArray<DoubleValue> adjQualities = new ItemArray<DoubleValue>(parentQualities.Length);

        for (int i = 0; i < adjQualities.Length; i++) {
          adjQualities[i] = new DoubleValue(parentQualities[i].Value - c * Math.Pow(parents[i].Size, k));
        }
        ParentAdjustedQualityParameter.ActualValue = adjQualities;
      } else {
        // adjusted fitness is equal to fitness
        ParentAdjustedQualityParameter.ActualValue = (ItemArray<DoubleValue>)ParentQualityParameter.ActualValue.Clone();
        OffspringAdjustedQualityParameter.ActualValue = (DoubleValue)OffspringQualityParameter.ActualValue.Clone();
      }
      return base.Apply();
    }
  }
}
